package main

import (
	"bytes"
	"math/rand"
	"strings"
	"testing"
)

const srcTxt = `Lorem ipsum dolor sit amet, consectetur adipiscing elit, sed do eiusmod tempor incididunt ut labore et dolore magna aliqua. Ut enim ad minim veniam, quis nostrud exercitation ullamco laboris nisi ut aliquip ex ea commodo consequat. Duis aute irure dolor in reprehenderit in voluptate velit esse cillum dolore eu fugiat nulla pariatur. Excepteur sint occaecat cupidatat non proident, sunt in culpa qui officia deserunt mollit anim id est laborum.`

func makeStringChunks() []string {
	// split the text into random sized rune chunks
	var runes []rune
	for _, r := range srcTxt {
		runes = append(runes, r)
	}

	var chunks []string
	for {
		l := len(runes)
		if l <= 0 {
			break
		}
		if l > 15 {
			l = 15
		}

		cl := rand.Intn(l) + 1
		chunk := runes[:cl]
		chunks = append(chunks, string(chunk))
		runes = runes[cl:]
	}

	return chunks
}

func makeByteSliceChunks() [][]byte {
	stringChunks := makeStringChunks()
	var chunks [][]byte
	for _, s := range stringChunks {
		chunks = append(chunks, []byte(s))
	}
	return chunks
}

func BenchmarkStringBuilder_String(b *testing.B) {
	chunks := makeStringChunks()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		var sb strings.Builder
		for _, chunk := range chunks {
			sb.WriteString(chunk)
		}
		if sb.String() != srcTxt {
			panic("resulting string differs")
		}
	}
}

func BenchmarkStringBuilder_ByteSlice(b *testing.B) {
	chunks := makeByteSliceChunks()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		var sb strings.Builder
		for _, chunk := range chunks {
			sb.Write(chunk)
		}
		if sb.String() != srcTxt {
			panic("resulting string differs")
		}
	}
}

func BenchmarkBytesBuffer_String(b *testing.B) {
	chunks := makeStringChunks()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		var bb bytes.Buffer
		for _, chunk := range chunks {
			bb.WriteString(chunk)
		}
		if bb.String() != srcTxt {
			panic("resulting string differs")
		}
	}
}

func BenchmarkBytesBuffer_ByteSlice(b *testing.B) {
	chunks := makeByteSliceChunks()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		var bb bytes.Buffer
		for _, chunk := range chunks {
			bb.Write(chunk)
		}
		if bb.String() != srcTxt {
			panic("resulting string differs")
		}
	}
}

func BenchmarkString_Concatenation(b *testing.B) {
	chunks := makeStringChunks()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		var s string
		for _, chunk := range chunks {
			s += chunk
		}
		if s != srcTxt {
			panic("resulting string differs")
		}
	}
}

func BenchmarkByteSlice_Append(b *testing.B) {
	chunks := makeByteSliceChunks()

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		var b []byte
		for _, chunk := range chunks {
			b = append(b, chunk...)
		}
		if string(b) != srcTxt {
			panic("resulting string differs")
		}
	}
}

func BenchmarkByteSlice_Append_Preallocated(b *testing.B) {
	chunks := makeByteSliceChunks()

	var cl int
	for _, chunk := range chunks {
		cl += len(chunk)
	}

	b.ResetTimer()
	for i := 0; i < b.N; i++ {
		b := make([]byte, 0, cl)
		for _, chunk := range chunks {
			b = append(b, chunk...)
		}
		if string(b) != srcTxt {
			panic("resulting string differs")
		}
	}
}
